"""
None of the existing empirical models perfectly describes the training data. More
specifically, at the end of the training, the measurement uncertainties do not explain
the dispersion of the residuals at the fitted model.
This means that there is variability from SN to SN,
bearing on the shape of the light curves, or on the spectral structure that cannot be
captured by a two parameter model.
It is therefore essential to model this intrinsic residual variability of SNe by an
error model. This one is an empirical function of the phase and the wavelength. It
depends on parameters that must be adjusted during the training procedure.

In this study, we have implemented a simple error model, as a proof of concept.
of concept. It includes a purely diagonal part (error snake), as well as a non-diagonal
part, which describes a specific error mode (color scatter) observed on the training.
"""

import numpy as np
import pylab as pl
from scipy.sparse import coo_matrix, dia_matrix, csc_matrix
from ..lib.bspline import BSpline, CardinalBSplineC, integ, BSpline2D, lgram
from ..lib.fitparameters import FitParameters
try:
    from sksparse.cholmod import cholesky_AAt
except ImportError:
    from scikits.sparse.cholmod import cholesky_AAt


class ColorScatter(object):
    r"""
    To model the residual variability of the SNe color not described by the model, we allow the relative amplitude of
    each light curve to vary by a quantity :math:`(1+kappa)`:

    .. math::
        \phi_{phot}(p_0) = X_0 (1+z) (1+kappa) \int S(\lambda, p) T\left( \lambda \right) \frac{\lambda}{hc} d\lambda

    where :math:`\kappa` is an additional parameter, depending on the SN and the observation band.

    The variability of :math:`\kappa` is defined by a centered Gaussian a priori and whose variance depends on the
    wavelength restframe, and must be determined during the adjustment.
    In our model, this term is implemented by a polynomial of the wavelength restframe of the SN. For a light curve
    observed in a band :math:`X`, of average wavelength :math:`\lambda^X`, the corresponding variance is:

    .. math::
        \sigma_\kappa^2 = P\left(\frac{\lambda^X}{1+z}\right)

    where :math:`z` is redshift of the supernova and :math:`P` is a polynomial of the wavelength restframe.
    In practice, :math:`P` is implemented in terms of a reduced restframe wavelength: :math:`\lambda_{r}`

    Attributes
    ----------
    WL_REDUCED : numpy.array
        Reduced SN-restframe mean wavelength
    pars : nacl.lib.fitparameters.FitParameters
        Color scatter parameter.
    """
    U_WAVELENGTH = 3650.88
    B_WAVELENGTH = 4302.57
    V_WAVELENGTH = 5428.55
    R_WAVELENGTH = 6418.01
    I_WAVELENGTH = 7968.34
    WAVELENGTH = {"U": U_WAVELENGTH, "B": B_WAVELENGTH, "V": V_WAVELENGTH,
                  "R": R_WAVELENGTH, "I": I_WAVELENGTH}

    def __init__(self, wavelength_rest_reduced, sigma_kappa):
        """
        Constructor - computes the arguments

        Parameters
        ----------
        wavelength_rest_reduced :
            Reduced SN-restframe mean wavelength
        sigma_kappa : numpy.array
            Color scatter parameter initialisation.
        """
        self.WL_REDUCED = wavelength_rest_reduced
        self.pars = self.init_pars(sigma_kappa)

    @staticmethod
    def init_pars(sigma_kappa):
        """
        Parameters initialisation.

        Parameters
        ----------
        sigma_kappa : numpy.array
            Color scatter parameter initialisation.

        Returns
        -------
        pars : nacl.lib.fitparameters.FitParameters
            Color scatter parameter.
        """
        pars = FitParameters([('sigma_kappa', len(sigma_kappa))])
        pars['sigma_kappa'] = sigma_kappa
        return pars
        
    def __call__(self, sk, jac=False):
        """
        Evaluate the color scatter.

        Parameters
        ----------
        sk : numpy.array
            Color scatter parameter

        Returns
        -------
        val : numpy.array
            Color scatter evaluation, square root of the diagonal term of the color scatter 
            matrix.
        if jac:
            jacobian of the color scatter.
        """
        self.pars.free = sk
        sigma_kappa = self.pars.full[:]
        val = np.polyval(sigma_kappa, self.WL_REDUCED)  # * self.WL_REDUCED
    
        if jac:
            vander = (np.vander(self.WL_REDUCED, len(sigma_kappa)).T )  # * self.WL_REDUCED)
            jj = (vander * val * 2).T
            return val, jj  # coo_matrix(jj)
        return val


class SimpleVarianceModel(object):
    r"""
    Implementation of the error snake,  residuals variability coming from SN Ia variability
    not modeled and instrumental uncertainty not took into account.
    The diagonal part of the error model (error snake) can be understood as the contribution
    of the higher order terms of the PCA, having been truncated.

    Variance model, function of the flux both for photometry and spectroscopy
    (indiscriminately represented by X)

    .. math::
    V(\lambda, p) = \sigma_{Err}^2 + \sigma_{\mathrm{Mod}}^2(\lambda,p)

    where :

    .. math::
        \sigma_{\mathrm{Mod}}^{spec} &= g(\lambda, p) \times \phi_{\mathrm{spec}}(p,\lambda)\ \ (\mathrm{spectra}) \\
        \sigma_{\mathrm{Mod}}^{phot} &= g(\lambda, p) \times \phi_{\mathrm{phot}}(p,\lambda)\ \ (\mathrm{photometric})

    :math:`\gamma(\lambda, p)` is a global surface describing the spectral residual of
    a 2D spline surface defined on the same ranges as the corresponding model.


    Attributes
    ----------
    disconnect_sp : bool
        Whether error snake of spectral data is considered.
    model : nacl.salt
        Model.
    sp : nacl.dataset.SpectrumData
        Spectral data
    lc : nacl.dataset.LcData
        Photometric data
    filter_db : nacl.instruments.FilterDb
        Filter transmission projected on B-splines
    bands : numpy.array
        Bands used for photometric data
    threshold : int
        Fix the spline parameters for bin that gather less data points than this threshold value.
    wl_grid : numpy.array
        Wavelength grid.
    phase_grid : numpy.array
        Phase grid.
    basis : nacl.lib.bspline.BSpline2D
        2D spline basis.
    factor_int : float
        Model normalisation.
    lambda_c : numpy.array
        Mean wavelength of the filter in the SN-restframe.
    pars : nacl.lib.fitparameters.FitParameters
        Parameters :math:`\gamma`.
    pars0 : numpy.array
        Parameters used to initialize the error snake.
    """
    def __init__(self, model, gamma_init=0.01,  disconnect_sp=False, threshold=None,
                 adaptative_grid=False, bins=(3, 3), order=4):
        r"""
        Constructor.
        - instantiate the bases (phase, wavelength)
        - compute the \lambda_c for each light curve
        - initiate the parameters
        

        Parameters
        ----------
        model : nacl.salt
            Model.
        gamma_init : numpy.array ou float
            Parameters used to initialize the error snake.
        disconnect_sp : bool
            Whether error snake of spectral data is considered.
        threshold : int
            Fix the spline parameters for bin that gather less data points than this threshold value.
        adaptative_grid : bool
            It true, define Bspline knots where parameter space has data.
        bins : tuple
            Number of bins if the grid is not adaptative.
        order :
            Spline order.
        """
        self.disconnect_sp = disconnect_sp
        
        self.model = model
        self.lc = self.model.training_dataset.lc_data
        self.sp = self.model.training_dataset.spec_data
        self.filter_db = self.model.filter_db
        self.bands = self.model.bands
        self.threshold = threshold
        
        ph_bin, wl_bin = bins
        order = order

        if adaptative_grid:
            phase_lc = self.model.get_restframe_phases(self.lc)
            phase_sp = self.model.get_restframe_phases(self.sp)
            wavelength_lc = self.lc['Wavelength']/(1 + self.lc['ZHelio'])
            wavelength_sp = self.sp['Wavelength']/(1 + self.sp['ZHelio'])
            phase = np.hstack((phase_sp, phase_lc))
            wavelength = np.hstack((wavelength_sp, wavelength_lc))
            _, self.wl_grid, self.phase_grid, _ = pl.hist2d(wavelength, phase, bins=2)

        else:
            
            print(f'\n VarModel N bins : ph : {ph_bin}, wl {wl_bin}')
            self.phase_grid = np.linspace(model.basis.by.range[0], model.basis.by.range[-1], ph_bin)
            self.wl_grid = np.linspace(model.basis.bx.range[0], model.basis.bx.range[-1], wl_bin)

        self.basis = BSpline2D(self.wl_grid, self.phase_grid,
                               x_order=order,
                               y_order=order)
        
        self.factor_int = self.model.norm
        if self.lc['Wavelength'].sum() == 0:
            filter_mean_wl = [self.model.filter_db.transmission_db[bd].mean() for bd in self.model.bands.astype(str)]
            self.lambda_c = np.array(filter_mean_wl)[self.lc['band_id']]/(1+self.lc['ZHelio']) 
        else:
            self.lambda_c = self.lc['Wavelength']/(1+self.lc['ZHelio'])

        self.pars = self.init_pars(gamma_init=gamma_init)
        self.pars0 = self.pars.full.copy()

    def fix_non_constrain_pars(self, threshold):
        """
        Return index of parameters of the parameter space that don't have enough data to constraint the error snake.

        Parameters
        ----------
        threshold : int
            Threshold value defining the "enough". If the sum of the data in a bin is less than this value,
            the corresponding parameter is fixed.

        Returns
        ----------
        idx : numpy.array
            Index of parameters that will be fixed.
        """
        lcs = self.model.training_dataset.lc_data
        sps = self.model.training_dataset.spec_data
        ph_sp = self.model.get_restframe_phases(sps)
        ph_lc = self.model.get_restframe_phases(lcs)
        wl_sp = sps['Wavelength']/(1 + sps['ZHelio'])
        wl_lc = lcs['Wavelength']/(1 + lcs['ZHelio'])

        x = np.hstack((wl_sp, wl_lc))
        y = np.hstack((ph_sp, ph_lc))
        jj = self.basis.eval(x, y)

        idx = np.where(np.bincount(np.sort(jj.col)) < threshold)
        return idx[0] 

    def init_pars(self, gamma_init):
        """
        Initiate model parameters.

        Parameters
        ----------
        gamma_init : float
            Parameters used to initialize the error snake.

        Returns
        ----------
        gamma : nacl.lib.fitparameters.FitParameters
            Parameters :math:`\gamma`.
        """
        gamma = FitParameters([('gamma', self.basis.bx.nj * self.basis.by.nj)])
        gamma['gamma'] = gamma_init
        if self.threshold is not None:
            idx = self.fix_non_constrain_pars(self.threshold)
            print(idx)
            print(gamma.free.shape)
            gamma.fix(idx)
            print(gamma.free.shape)
        return gamma
                    
    def __call__(self, gamma, beta, jacobian=None, debug_mode=False, lc=None, sp=None):
        r"""
        Evaluate the model for a set of parameter :math:`\gamma` (variance model parameters)
        and :math:`\beta` (flux model parameters).
        

        Parameters
        ----------
        gamma : np.array
           set of parameter of the variance model;
        beta : np.array
           set of parameter of the flux model;
        jacobian : None or scipy.sparse.coo_matrix
           derivative of the flux, not necessary if derivative of variance model not needed,
        debug_mode : bool
           to get some variables...
        lc : nacl.dataset.LcData
           to get variance model of a set of light curve and is None will get for all
        sp : nacl.dataset.SpectrumData
           to get variance model of a set of spectra and is None will get for all

        Returns
        ----------
        var : numpy.array
            Variance value of the model, quadratic sum of the error measurement and the error snake.
        if jacobian is not None:
        var : numpy.array
            Variance value of the model, quadratic sum of the error measurement and the error snake.
        jacobian_var : scipy.sparse.coo_matrix
            Jacobian (derivatives wrt :math:`\beta` and :math:`\gamma`)matrix of the error model.
        """
        if lc is None:
            lc = self.lc
        if sp is None:
            sp = self.sp
        lambda_c = self.lambda_c[lc['i']]

        self.model.pars.free = beta
        flux = self.model(beta, jac=False)
        self.pars.full[:] = gamma

        phase_lc = (lc['Date']-self.model.pars['tmax'].full[lc['sn_id']])/(1+lc['ZHelio'])
        phase_sp = (sp['Date']-self.model.pars['tmax'].full[sp['sn_id']])/(1+sp['ZHelio'])

        if self.disconnect_sp:
            sigma_spectra = np.zeros_like(phase_sp)
        else:
            wl_sp = np.array(sp['Wavelength'])/(1+sp['ZHelio'])
            spline_sp = self.basis.eval(wl_sp, phase_sp)
            flux_sp = flux[sp['i']]
            ss_sp = (spline_sp * self.pars['gamma'].full)
            sigma_spectra = ss_sp * flux_sp

        spline_lc = self.basis.eval(lambda_c, phase_lc)
        flux_lc = flux[lc['i']] 
        ss_lc = (spline_lc * self.pars['gamma'].full)
        sigma_lightcurve = ss_lc * flux_lc
        sig = np.hstack((sigma_lightcurve, sigma_spectra))
        flux_err = np.hstack((lc['FluxErr'], sp['FluxErr']))
        var = sig**2 + flux_err**2 
        
        if jacobian is not None:
            n_pars, n_data = len(self.model.pars.free) + len(self.pars.free), len(sig)
            v, jacobian = self.model(beta, jac=True)

            if self.disconnect_sp:
                i_dp, j_dp, jacobian_pars, i_dg0, j_dg0, jacobian_gamma = [], [], [], [], [], []
            else:
                jacobian_spectra = jacobian.tocsc()[sp['i']].tocoo()
                # dp
                i_dp = jacobian_spectra.row.copy()
                j_dp = jacobian_spectra.col.copy()
                jacobian_pars = 2 * jacobian_spectra.data * (sigma_spectra * ss_sp)[i_dp]
                
                # dtmax
                idx = self.model.pars['tmax'].indexof(sp['sn_id'])
                jacobian_wrt_phase, jacobian_wrt_wavelength = self.basis.gradient(wl_sp,
                                                                                  phase_sp + self.model.delta_phase)
                if len(jacobian_pars) != 0.:
                    jacobian_pars[idx] -= 2 * (jacobian_wrt_phase * self.pars['gamma'].full) / \
                                          (1+sp['ZHelio']) * v[sp['i']] * sigma_spectra
                # dg
                i_dg0 = spline_sp.row.copy()
                j_dg0 = self.pars['gamma'].indexof(spline_sp.col) + len(self.model.pars.free)
                jacobian_gamma = 2 * v[sp['i']][i_dg0]**2 * spline_sp.data * ss_sp[i_dg0]
                if debug_mode:
                    return i_dp, j_dp, jacobian_pars, i_dg0, j_dg0, jacobian_gamma, v, sp, spline_sp, ss_sp

            v_lc = v[lc['i']]
            jacobian_lightcurve = jacobian.tocsc()[lc['i']].tocoo()
            
            i_dp_lc = jacobian_lightcurve.row.copy()
            j_dp_lc = jacobian_lightcurve.col.copy()
            jacobian_pars_lightcurve = 2 * jacobian_lightcurve.data * (sigma_lightcurve * ss_lc)[i_dp_lc]
            
            # dtmax
            if self.model.pars['tmax'].indexof().sum() != len(self.model.pars['tmax'].full) * -1:
                jacobian_wrt_phase, jacobian_wrt_wavelength = self.basis.gradient(lambda_c,
                                                                                        phase_lc +
                                                                                        self.model.delta_phase)
                idx_lc = np.where(np.in1d(jacobian_lightcurve.col, self.model.pars['tmax'].indexof()))
                if len(jacobian_pars_lightcurve) != 0.:
                    jacobian_pars_lightcurve[idx_lc] -= 2 * (jacobian_wrt_phase * self.pars['gamma'].full) / (1+lc['ZHelio']) * v_lc * sigma_lightcurve
                
            # dg
            i_dg0_lc = spline_lc.row.copy()
            j_dg0_lc = self.pars['gamma'].indexof(spline_lc.col) + len(self.model.pars.free)
            jacobian_gamma_lightcurve = 2 * v_lc[i_dg0_lc] ** 2 * spline_lc.data * ss_lc[i_dg0_lc]
            
            if len(i_dg0) != 0:
                i_dg0 += len(lc)
            if len(i_dp) != 0:
                i_dp += len(lc)
            
            i = np.hstack((i_dp, i_dg0, i_dp_lc, i_dg0_lc))
            j = np.hstack((j_dp, j_dg0, j_dp_lc, j_dg0_lc))
            val = np.hstack((jacobian_pars, jacobian_gamma, jacobian_pars_lightcurve, jacobian_gamma_lightcurve))
             
            idx_zeros = np.where((j != -1))[0]
            jacobian_var = coo_matrix((val[idx_zeros], (i[idx_zeros], j[idx_zeros])), shape=(n_data, n_pars))

            return var, jacobian_var.T
        return var





import setuptools
from setuptools import setup, Extension


setup(name='sn-nacl',
      #      packages=setuptools.find_packages(),
      packages=['nacl', 'nacl.lib', 'nacl.models',
                'nacl.plotting', 'nacl.simulations', 'nacl.util'],
      ext_modules=[Extension('nacl.lib._libnacl',
                             ['nacl/lib/utils.c',],
                             extra_compile_args=['-std=c11', '-O3'])],
      version="0.3",
      description="Supernova empirical model training package",
      author='G. Augarde, N. Regnault, M. Betoule, S. Bongard',
      author_email='nicolas.regnault@lpnhe.in2p3.fr',
      #      package_dir={'': 'nacl'},
      zip_safe=False,
      
      install_requires=['numpy>=1.7', 'matplotlib', 'seaborn', 'scikit-sparse', 'astropy', 'sparseqr'],
      setup_requires=['pytest-runner'],
      tests_require=['pytest-3'],
)



"""
The SALT2 model. 

SALT2 parameterizes the *restframe* spectrophotometric evolution of
SNe Ia with two surfaces and a color law:

.. math:: S_{10}(p, \lambda) = X_0 \\left(M_0(p, \lambda) + X_1 M_1(p,\lambda)\\right) C(\lambda)

Given a passband, :math:`T(\lambda)`, the model allows one to compute
an *observer frame* broadband flux:

.. math:: f = \\frac{C\ 10^{-12}}{1+z} \int S_{10}\\left(p, \\frac{\lambda}{1+z}\\right)\ \lambda\ T(\lambda)\ d\lambda

"""

import os
import os.path as op
import sys 
import re

import numpy as np 
from scipy.sparse import coo_matrix

from ..util.io import NTuple
from ..lib.stellarlibs import FilterSet, ALIGN_WITH_INTEG_FLUX
from ..lib.instruments import InstrumentModel 
from ..lib.fitparameters import FitParameters
from ..lib.bspline import BSpline, BSpline2D, lgram
from . import saltpath


def load_filters(bands):
    """
    Load all the filters that are requested in argument and return a
    FilterSet
    
    .. note :: this function is currently slow, because no instrument
               cache implemented.
    """
    instrument_names = np.unique([re.search('(.+)::(.+)', bn).group(1) for bn in bands])
    instruments = [InstrumentModel(nm) for nm in instrument_names]
    instruments = dict(list(zip(instrument_names, instruments)))
    tr, lmin, lmax = [], [], []
    for bn in bands:
        instr, band = re.search('(.+)::(.+)', bn).groups()
        f = instruments[instr].EffectiveFilterByBand(band)
        lmin.append(f.x_min)
        lmax.append(f.x_max)
        tr.append(f)
    gwl = np.arange(min(2000., np.min(lmin)), 
                    max(11000., np.max(lmax)), 5.)
    basis = BSpline(gwl, order=2)
    return FilterSet(basis, tr)


class ModelComponents(object):
    """Basic ingredients of the SALT2 model
    
    Holds all the model ingredients that are shared between all supernovae:
      - model validity area in phase and wavelength
      - BSpline basis definitions
      - surfaces
      - color law
      - error snakes
      - color smearing laws
    """
    def __init__(self, filename=None):
        self.basis, self.wl_basis, self.phase_basis = None, None, None
        self.M0, self.M1 = None, None
        self.cl = None
        if filename is not None:
            self.load(filename)

    def save(self, filename):
        if self.basis is None:
            return 
        np.savez(filename, 
                 wl_grid=self.wl_basis.grid,
                 phase_grid=self.phase_basis.grid, 
                 M0=self.M0, 
                 M1=self.M1, 
                 CL_pars=self.cl.pars, 
                 CL_range=[self.cl.min_lambda, self.cl.max_lambda])
    
    def load(self, filename):
        arrs = np.load(filename)
        wl_grid, phase_grid = arrs['wl_grid'], arrs['phase_grid']
        self.basis = BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)
        self.wl_basis = self.basis.bx
        self.phase_basis = self.basis.by
        self.M0, self.M1 = arrs['M0'], arrs['M1']
        self.cl = ColorLaw(arrs['CL_pars'], arrs['CL_range'])

    @staticmethod
    def load_from_saltpath():
        """load the SALT2 model from an "official" SALT2 repository
        """

        phase_grid = np.linspace(-20., 51., 50)
        wl_grid    = np.linspace(2000., 9201., 300)
        basis      = BSpline2D(phase_grid, wl_grid, x_order=4, y_order=4)
        
        cfg = saltpath.read_card_file(saltpath.fitmodel_filename)
        salt2_model_path = saltpath.SALTPATH + os.sep + cfg['SALT2']
    
        # first surface
        M0_filename = salt2_model_path + os.sep + 'salt2_template_0.dat'
        d = NTuple.fromtxt(M0_filename)
        phase, wl, flx = d['f0'].copy(), d['f1'].copy(), d['f2'].copy()
        M0 = basis.linear_fit(phase, wl, flx)
    
        # second surface 
        M1_filename = salt2_model_path + os.sep + 'salt2_template_1.dat'
        d = NTuple.fromtxt(M1_filename)
        phase, wl, flx = d['f0'].copy(), d['f1'].copy(), d['f2'].copy()
        M1 = basis.linear_fit(phase, wl, flx)
    
        # color law 
        print((salt2_model_path + os.sep + 'salt2_color_correction.dat'))
        cl_filename = salt2_model_path + os.sep + 'salt2_color_correction.dat'
        with open(cl_filename) as f:
            deg = int(f.readline())
            p = []
            for i in range(deg):
                p.append(float(f.readline()))
            r = {}
            for line in f:
                field, value = re.search('^Salt2ExtinctionLaw\.([\w_]+)\s*([\d\.+-]+)', line).groups()
                r[field] = float(value)
            assert 'version' in r and r['version'] == 1.0
        
        ret = ModelComponents()
        ret.basis = basis
        ret.phase_basis = basis.bx
        ret.wl_basis = basis.by
        ret.M0 = M0.reshape(len(ret.wl_basis), -1)
        ret.M1 = M1.reshape(len(ret.wl_basis), -1)
        alpha = 1. - np.sum(p)
        p = np.asarray(p[::-1] + [alpha])
        ret.cl = ColorLaw(p, (r['min_lambda'], r['max_lambda']))
        ret.CC = ret.cl(ret.wl_basis.grid, 1.)
        return ret


class ColorLaw(object):
    """Implementation of the SALT2 color law (version 1)
    
    The SALT2 color law describes the color diversity of SNeIa. It is
    fitted from a training sample. In SALT 2.4, the color law is
    parametrized as follows:
    
    .. math:: C(\lambda;c) = 10^{0.4 (c \times P(\lambda))}

    where :math:`P(\lambda)` is a polynomial, such as
    :math:`P(\lambda_B) = 1` and :math:`P(\lambda_V) = 0` It is
    defined in the wavelength range :math:`2800 A < \lambda < 7000 A`
    """
    WAVELENGTH = {"U": 3650.88, "B": 4302.57, "V": 5428.55, "R": 6418.01, "I": 7968.34}
    U_WAVELENGTH=3650.88
    B_WAVELENGTH=4302.57
    V_WAVELENGTH=5428.55
    R_WAVELENGTH=6418.01
    I_WAVELENGTH=7968.34
    
    def __init__(self, pars, wl_range):
        """Constructor - computes the arguments 
        
        """
        assert(wl_range[0] < wl_range[1])
        self.min_lambda, self.max_lambda = wl_range
        
        self.pars = pars
        
        self.min_reduced_lambda = self.reduce(self.min_lambda)
        self.max_reduced_lambda = self.reduce(self.max_lambda)
        
        dpars = np.polyder(self.pars)
        
        # uv side 
        val = np.polyval(self.pars, self.min_reduced_lambda)
        dval = np.polyval(dpars, self.min_reduced_lambda)
        self.pars_uv = np.asarray([dval * self.min_reduced_lambda + val,
                                   val * self.min_reduced_lambda])
        # ir side 
        val = np.polyval(self.pars, self.max_reduced_lambda)
        dval = np.polyval(dpars, self.max_reduced_lambda)
        self.pars_ir = np.asarray([dval * self.max_reduced_lambda + val,
                                   val * self.max_reduced_lambda])
        
    def reduce(self, wl):
        B_WL, V_WL = self.WAVELENGTH["B"], self.WAVELENGTH["V"]
        return (wl-B_WL)/(V_WL-B_WL);
    
    def __call__(self, wl, color):
        rwl = self.reduce(np.asarray(wl))
        r = np.polyval(self.pars, rwl) * rwl
        idx_uv = rwl<self.min_reduced_lambda
        r[idx_uv] = np.polyval(self.pars_uv, rwl[idx_uv]-self.min_reduced_lambda)
        idx_ir = rwl>self.max_reduced_lambda
        r[idx_ir] = np.polyval(self.pars_ir, rwl[idx_ir]-self.max_reduced_lambda)
        return np.power(10., 0.4 * color * r)


def precompute_filter_effective_restframe_lambda(model, fs, zgrid):
    """re
    """
    def effective_lambda(model, fs, z):
        """
        Compute 
        \int \lambda^2 B_i(\lambda) B_j(\lambda (1+z)) d\lambda / \int \lambda B_i B_j d\lambda
        """
        G  = coo_matrix(lgram(model.wl_basis, fs.basis, z=z))
        G2 = coo_matrix(lgram(model.wl_basis, fs.basis, z=z, lambda_power=2))
        assert ~np.any(G.row-G2.row) and ~np.any(G.col-G2.col)
        leff = G2.data/G.data
        L = coo_matrix((leff, (G.row, G.col)), shape=G.shape)
        N = coo_matrix((np.ones_like(leff), (G.row, G.col)), shape=G.shape)
        leff = np.array(L.sum(axis=1)).squeeze()
        N = np.array(N.sum(axis=1)).squeeze()
        idx = N>0
        leff[idx] /= N[idx]
        return leff
        
    #    self.CC = self.model.cl(self.leff / (1.+self.z), 1.)
    #    self.CC[~idx] = 1.
    leff = []
    # refine the z_grid
    for z in zgrid:
        leff.append(effective_lambda(model, fs, z))
        print((z, leff[-1][0], leff[-1][-1]))
    leff = np.vstack(leff)
    bs = BSpline2D(model.wl_basis.grid, zgrid, x_order=1, y_order=1)
    X,Y = np.meshgrid(model.wl_basis.grid, zgrid)
    
    return bs, X.flatten(), Y.flatten(), leff


    
def param_setter(name):
    def set_param(self, val):
        self.pars[name] = val
    return set_param


def param_getter(name):
    def get_param(self):
        return self.pars[name].full[0]
    return get_param


class SALT2(object):
    """Evaluate the SALT2 model for a specific supernova

    An instance of this class takes as an input a list of observations
    (MJDs, bands).  It precomputes and caches everything that can be
    pre-computed, essentially, the color law, the gram at that
    redshift and the Gram folded with the passband coefficients.  

    When __call__'ed with a parameter vector, it evaluates the SALT2
    model for each observation.
    
    """
    def __init__(self, mjds, bands, model, filterset, z=0., wl_range=(3000., 8000.)):
        """Constructor

        Stores the list of observations, initialize a parameter vector, 
        
        """
        # basic stuff 
        self.mjds = mjds
        self.bands = bands
        #        self.bands_unique, self.band_index =
        self.build_band_index(bands)
        self.model = model
        # assumes filterset has all bands that are needed
        self.filterset = filterset
        # expect M0 and M1 to be matrices
        self.M0, self.M1 = model.M0, model.M1
        self.restframe_wavelength_range = wl_range
        
        # set up an internal parameter vector 
        self.pars = self.init_pars()
        
        # set the redshift. This triggers a recompute 
        # of the Gram and Fz matrices and the color law
        self.z = z
        
    # properties, for easy access to the internal parameter vector (somewhat slow)
    X0 = property(fget=param_getter('X0'), fset=param_setter('X0'))
    X1 = property(fget=param_getter('X1'), fset=param_setter('X1'))
    Color = property(fget=param_getter('Color'), fset=param_setter('Color'))
    DayMax = property(fget=param_getter('DayMax'), fset=param_setter('DayMax'))

    # special property for z, because when we change z, 
    # we have to recompute the grams and the color law...
    @property
    def z(self):
        return self.pars['z'].full[0]
    @z.setter
    def z(self, val):
        self.pars['z'] = val
        self.G, self.Fz = self.init_fz()
        self.init_CC()

    def build_band_index(self, bands):
        self.bands_unique = np.unique(bands)
        idx = dict(list(zip(self.bands_unique, np.arange(len(self.bands_unique)))))
        self.band_index = np.array([idx[b] for b in bands])
        return self.bands_unique, self.band_index        
        
    def init_fz(self):
        """creates and return the F(z) matrix
        """
        G = lgram(self.model.wl_basis, self.filterset.basis, z=self.z)
        Fz = np.vstack([G * self.filterset[b] for b in self.bands_unique])
        return G, Fz
    
    def init_CC(self):
        """evaluate the color law
        """
        if not hasattr(self, 'G'):
            self.G = lgram(self.model.wl_basis, self.filterset.basis, z=self.z)
        G = coo_matrix(self.G)
        self.G2 = lgram(self.model.wl_basis, self.filterset.basis, z=self.z, lambda_power=2)
        G2 = coo_matrix(self.G2)
        assert ~np.any(G.row-G2.row) and ~np.any(G.col-G2.col)
        leff = G2.data/G.data
        L = self.L = coo_matrix((leff, (G.row, G.col)), shape=G.shape)
        N = self.N = coo_matrix((np.ones_like(leff), (G.row, G.col)), shape=G.shape)
        self.leff = np.array(L.sum(axis=1)).squeeze()
        N = np.array(N.sum(axis=1)).squeeze()
        idx = N>0
        self.leff[idx] /= N[idx]
        self.CC = self.model.cl(self.leff / (1.+self.z), 1.)
        self.CC[~idx] = 1.

    def mask_forbidden_bands(self):
        """evaluate the mask
        """
        mean_restframe_wavelength = self.filterset.mean_wavelength(self.bands) / (1.+ self.z)
        blue_cutoff, red_cutoff = self.restframe_wavelength_range
        return (mean_restframe_wavelength>blue_cutoff) & (mean_restframe_wavelength<red_cutoff)
        
    @staticmethod
    def init_pars():
        p = FitParameters([('X0', 1), ('X1', 1), ('Color', 1), ('DayMax', 1), ('z', 1)])
        p['X0'] = 1.
        p['X1'] = 0.
        p['Color'] = 0.
        p['DayMax'] = 0.
        p['z'] = 0.
        p.fix('z')
        return p
    
    def __call__(self, p=None, jac=False):
        """evaluate the model and its derivatives (optional)
        
        The broadband flux is given by:

        .. math:: F(p) = X_0 \times \\left(I_0 + X_1 I_1\\right)) \\times  10^{0.4 c P(\lambda)}
        """
        if p is not None:
            self.pars.free = p 
        M0, M1, Fz = self.M0, self.M1, self.Fz
        z, X0, X1, Color, DayMax = self.z, self.X0, self.X1, self.Color, self.DayMax
        
        phase = (self.mjds-DayMax) / (1. + self.z)

        # using a sparse csc() makes a noticeable difference
        # (x4 in speed with respect to multiply dense matrices)
        # even in the other matrix (M0, M1) is dense
        # (M0, M1 has to be a matrix though, not an array)
        # 
        # *TODO*: understand what underlying function is called
        # 
        P = self.model.phase_basis.eval(phase).tocsc()
        CC = np.power(self.CC, Color)
        M0_CC  = np.matrix(self.M0.T * CC)
        M1_CC  = np.matrix(self.M1.T * CC)
        Fz = Fz[self.band_index]
        I0 = (np.array(P * M0_CC) * Fz).sum(axis=1)
        I1 = (np.array(P * M1_CC) * Fz).sum(axis=1)
        
        # OLD version 
        #        # seems to be much faster with dense matrices
        #        P = self.model.phase_basis.eval(phase).todense()
        #        CC = np.power(self.CC, Color)
        #        I0 = (np.array(P * (M0.T * CC)) * Fz).sum(axis=1)
        #        I1 = (np.array(P * (M1.T * CC)) * Fz).sum(axis=1)        
        
        # I add the (1+z) factor and the ALIGN_WITH_INTEG_FLUX here
        norm = 1.E-12 * (1.+z) * ALIGN_WITH_INTEG_FLUX
        flux = norm * X0 * (I0 + X1*I1)
        
        # no derivatives ? just return the model value 
        if not jac:
            return flux
        
        # compute the derivatives w.r.t. parameters 
        n, N = len(self.pars.free), len(phase)
        J = np.zeros((N,n))
        
        # ddX0
        i = self.pars['X0'].indexof(0)
        if i>=0:
            J[:, i] = norm * (I0 + X1*I1)
            
        # ddX1
        i = self.pars['X1'].indexof(0)
        if i>=0:
            J[:, i] = norm * (X0 * I1)
        
        # ddDayMax 
        i = self.pars['DayMax'].indexof(0)
        if i>=0:
            #            dP = self.model.phase_basis.deriv(phase).todense() * (-1. / (1.+z))
            dP = self.model.phase_basis.deriv(phase)
            dP.data *= (-1. / (1.+z))
            dP = dP.tocsc()
            dI0 = (np.array(dP * M0_CC) * Fz).sum(axis=1)
            dI1 = (np.array(dP * M0_CC) * Fz).sum(axis=1)
            J[:, i] = norm * X0 * (dI0 + X1 * dI1)
        
        # ddColor 
        i = self.pars['Color'].indexof(0)
        if i>=0:
            dCC = np.log(self.CC) * CC
            M0_dCC, M1_dCC = np.matrix(M0.T * dCC), np.matrix(M1.T * dCC)
            dI0 = (np.array(P * M0_dCC) * Fz).sum(axis=1)
            dI1 = (np.array(P * M1_dCC) * Fz).sum(axis=1)
            J[:, i] = norm * X0 * (dI0 + X1 * dI1)
        
        # ddz - this one is costly to compute
        # I don't know how to compute it other than numerically
        i = self.pars['z'].indexof(0)
        if i>=0:
            self.z += dz
            dflux = self()
            J[:,i] = (dflux-flux)/dz
            self.z -= dz
        
        return flux, J
    
    def deriv_debug(self, p=None):
        if p is not None:
            self.pars.free = p 
        
        flx0 = self(jac=0)

        n, N = len(self.pars.free), len(self.mjds)
        J = np.zeros((N,n))        
        
        # ddX0
        self.X0 += 0.01
        flx = self()
        J[:, self.pars['X0'].indexof(0)] = (flx-flx0) / 0.01
        self.X0 -= 0.01
        # ddX1
        self.X1 += 0.01
        flx = self()
        J[:, self.pars['X1'].indexof(0)] = (flx-flx0) / 0.01
        self.X1 -= 0.01
        # ddDayMax 
        self.DayMax += 0.01
        flx = self()
        J[:, self.pars['DayMax'].indexof(0)] = (flx-flx0) / 0.01
        self.DayMax -= 0.01        
        # ddColor 
        self.Color += 0.001
        flx = self()
        J[:, self.pars['Color'].indexof(0)] = (flx-flx0) / 0.001
        self.Color -= 0.001      
        # ddz
        self.z += 0.1
        flx = self()
        J[:, self.pars['z'].indexof(0)] = (flx-flx0) / 0.01
        self.z -= 0.1

        return J 
        
def test():
    r = []
    for sn in sne:
        r.append(SALT2(sn))




